import pandas as pd
from tqdm import tqdm
def valueCounts(data,id_name,y_name=None):
    '''
        输入数据以及主键、y值(有就输入，没有只要输入data和id_name就可以)
        -------
        输出两个数据集，一个是字符变量的：
    	value	counts	varname
        0	S	644	Embarked
        1	C	168	Embarked
        2	Q	77	Embarked
        一个是数值型的：
        	        count	   mean	                  std	        min	 25%	 50%	75%	      max
            Pclass	891.0	2.308641975308642	0.8360712409770513	1.0	 2.0    3.0	    3.0	     3.0
            Age	    714.0	29.69911764705882	14.526497332334044	0.42 20.125	28.0	38.0	80.0
            SibSp	891.0	0.5230078563411896	1.1027434322934275	0.0	0.0	    0.0	    1.0	    8.0
            Parch	891.0	0.38159371492704824	0.8060572211299559	0.0	0.0	    0.0	    0.0	    6.0
            Fare	891.0	32.2042079685746	49.693428597180905	0.0	7.9104	14.4542	31.0	512.3292
    '''
    if  y_name==None:
        data_number=pd.DataFrame()
        data_char=pd.DataFrame()
        for k in tqdm(data.drop([id_name],axis=1).columns):
            if "int" in str(data[k].dtypes ) or "float" in str(data[k].dtypes):
                gg=data[k].describe().reset_index(drop=False)
                for l in gg["index"]:
                    data_number.loc[k,l]=gg[gg["index"]==l][k].iloc[0]
            elif "object" in str(data[k].dtypes):
                gg1=data[k].value_counts().reset_index(drop=False).rename(columns={"index":"value",k:"counts"})
                gg1["varname"]=k
                data_char=data_char.append(gg1)
    else:
        data_number=pd.DataFrame()
        data_char=pd.DataFrame()
        for k in tqdm(data.drop([id_name,y_name],axis=1).columns):
            if "int" in str(data[k].dtypes ) or "float" in str(data[k].dtypes):
                gg=data[k].describe().reset_index(drop=False)
                for l in gg["index"]:
                    data_number.loc[k,l]=gg[gg["index"]==l][k].iloc[0]
                data_number.loc[k,"缺失比例"]=1-gg[gg["index"]=="count"][k].iloc[0]/len(data)
            elif "object" in str(data[k].dtypes):
                gg1=data[k].value_counts().reset_index(drop=False).rename(columns={"index":"value",k:"counts"})
                gg1["varname"]=k
                gg1["缺失比例"]=1-data[k].count()/len(data)
                data_char=data_char.append(gg1)    
    return data_number,data_char
if __name__ == '__main__':
    train_data = pd.read_csv(r"F:\论文-风控\train.csv")  #读取训练集
    data1,data2=valueCounts(train_data,"id",y_name="isDefault")







#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Sep  7 15:22:54 2019

@author: betty
"""

'''
卡方分组
'''
import pandas as pd
def Chi2(df, total_col, bad_col):
    '''
    :param df: 包含全部样本总计与坏样本总计的数据框
    :param total_col: 全部样本的个数
    :param bad_col: 坏样本的个数
    :return: 卡方值
    '''
    df2 = df.copy()
    # 求出df中，总体的坏样本率和好样本率
    badRate = sum(df2[bad_col])*1.0/sum(df2[total_col])
        # print(badRate)
    df2['good'] = df2.apply(lambda x: x[total_col] - x[bad_col], axis = 1)
    goodRate = sum(df2['good']) * 1.0 / sum(df2[total_col])
    # 期望坏（好）样本个数＝全部样本个数*平均坏（好）样本占比
    df2['badExpected'] = df[total_col].apply(lambda x: x*badRate)
    df2['goodExpected'] = df[total_col].apply(lambda x: x * goodRate)
    badCombined = zip(df2['badExpected'], df2[bad_col])
    goodCombined = zip(df2['goodExpected'], df2['good'])
    badChi = [(i[0]-i[1])**2/i[0] for i in badCombined]
    goodChi = [(i[0] - i[1]) ** 2 / i[0] for i in goodCombined]
    chi2 = sum(badChi) + sum(goodChi)
    return chi2


def BinBadRate(df, col, target, grantRateIndicator=0):
    '''
    :param df: 需要计算好坏比率的数据集
    :param col: 需要计算好坏比率的特征
    :param target: 好坏标签
    :param grantRateIndicator: 1返回总体的坏样本率，0不返回
    :return: 每箱的坏样本率，以及总体的坏样本率（当grantRateIndicator＝＝1时）
    '''
    total = df.groupby([col])[target].count()
    total = pd.DataFrame({'total': total})
    bad = df.groupby([col])[target].sum()
    bad = pd.DataFrame({'bad': bad})
    regroup = total.merge(bad, left_index=True, right_index=True, how='left')
    regroup.reset_index(level=[0, ],inplace=True)
    regroup['bad_rate'] = regroup.apply(lambda x: x.bad * 1 / x.total, axis=1)
    dicts = dict(zip(regroup[col],regroup['bad_rate']))
    if grantRateIndicator==0:
        return (dicts, regroup)
    N = sum(regroup['total'])
    B = sum(regroup['bad'])
    overallRate = B * 1.0 / N
    return (dicts, regroup, overallRate)

### ChiMerge_MaxInterval: split the continuous variable using Chi-square value by specifying the max number of intervals
def ChiMerge(df, col, target, max_interval=5,special_attribute=[],minBinPcnt=0):
    '''
    :param df: 包含目标变量与分箱属性的数据框
    :param col: 需要分箱的属性
    :param target: 目标变量，取值0或1
    :param max_interval: 最大分箱数。如果原始属性的取值个数低于该参数，不执行这段函数
    :param special_attribute: 不参与分箱的属性取值
    :param minBinPcnt：最小箱的占比，默认为0
    :return: 分箱结果
    '''
    colLevels = sorted(list(set(df[col])))
    N_distinct = len(colLevels)
    if N_distinct <= max_interval:  #如果原始属性的取值个数低于max_interval，不执行这段函数
        print ("The number of original levels for {} is less than or equal to max intervals".format(col))
        return colLevels[:-1]
    else:
        if len(special_attribute)>=1:
            df1 = df.loc[df[col].isin(special_attribute)]
            df2 = df.loc[~df[col].isin(special_attribute)]
        else:
            df2 = df.copy()
        N_distinct = len(list(set(df2[col])))

        # 步骤一: 通过col对数据集进行分组，求出每组的总样本数与坏样本数
        if N_distinct > 100:
            df2=df2.sort_values(col)
            split_x = SplitData(df2, col, 100)
            df2['temp'] = df2[col].map(lambda x: AssignGroup(x, split_x))
        else:
            df2['temp'] = df2[col]
        # 总体bad rate将被用来计算expected bad count
        (binBadRate, regroup, overallRate) = BinBadRate(df2, 'temp', target, grantRateIndicator=1)
        

        # 首先，每个单独的属性值将被分为单独的一组
        # 对属性值进行排序，然后两两组别进行合并
        colLevels = sorted(list(set(df2['temp'])))
        groupIntervals = [[i] for i in colLevels]

        # 步骤二：建立循环，不断合并最优的相邻两个组别，直到：
        # 1，最终分裂出来的分箱数<＝预设的最大分箱数
        # 2，每箱的占比不低于预设值（可选）
        # 3，每箱同时包含好坏样本
        # 如果有特殊属性，那么最终分裂出来的分箱数＝预设的最大分箱数－特殊属性的个数
        split_intervals = max_interval - len(special_attribute)
        while (len(groupIntervals) > split_intervals):  # 终止条件: 当前分箱数＝预设的分箱数
            # 每次循环时, 计算合并相邻组别后的卡方值。具有最小卡方值的合并方案，是最优方案
            chisqList = []
            for k in range(len(groupIntervals)-1):
                temp_group = groupIntervals[k] + groupIntervals[k+1]
                df2b = regroup.loc[regroup['temp'].isin(temp_group)]
                #chisq = Chi2(df2b, 'total', 'bad', overallRate)
                chisq = Chi2(df2b, 'total', 'bad')
                chisqList.append(chisq)
            best_comnbined = chisqList.index(min(chisqList))
            groupIntervals[best_comnbined] = groupIntervals[best_comnbined] + groupIntervals[best_comnbined+1]
            # after combining two intervals, we need to remove one of them
            groupIntervals.remove(groupIntervals[best_comnbined+1])
        groupIntervals = [sorted(i) for i in groupIntervals]
        cutOffPoints = [max(i) for i in groupIntervals[:-1]]
        # 检查是否有箱没有好或者坏样本。如果有，需要跟相邻的箱进行合并，直到每箱同时包含好坏样本
        groupedvalues = df2['temp'].apply(lambda x: AssignBin(x, cutOffPoints))
        df2['temp_Bin'] = groupedvalues
        (binBadRate,regroup) = BinBadRate(df2, 'temp_Bin', target)
        [minBadRate, maxBadRate] = [min(binBadRate.values()),max(binBadRate.values())]
        while minBadRate ==0 or maxBadRate == 1:
            # 找出全部为好／坏样本的箱
            indexForBad01 = regroup[regroup['bad_rate'].isin([0,1])].temp_Bin.tolist()
            bin=indexForBad01[0]
            # 如果是最后一箱，则需要和上一个箱进行合并，也就意味着分裂点cutOffPoints中的最后一个需要移除
            if bin == max(regroup.temp_Bin):
                cutOffPoints = cutOffPoints[:-1]
            # 如果是第一箱，则需要和下一个箱进行合并，也就意味着分裂点cutOffPoints中的第一个需要移除
            elif bin == min(regroup.temp_Bin):
                cutOffPoints = cutOffPoints[1:]
            # 如果是中间的某一箱，则需要和前后中的一个箱进行合并，依据是较小的卡方值
            else:
                # 和前一箱进行合并，并且计算卡方值
                currentIndex = list(regroup.temp_Bin).index(bin)
                prevIndex = list(regroup.temp_Bin)[currentIndex - 1]
                df3 = df2.loc[df2['temp_Bin'].isin([prevIndex, bin])]
                (binBadRate, df2b) = BinBadRate(df3, 'temp_Bin', target)
                #chisq1 = Chi2(df2b, 'total', 'bad', overallRate)
                chisq1 = Chi2(df2b, 'total', 'bad')
                # 和后一箱进行合并，并且计算卡方值
                laterIndex = list(regroup.temp_Bin)[currentIndex + 1]
                df3b = df2.loc[df2['temp_Bin'].isin([laterIndex, bin])]
                (binBadRate, df2b) = BinBadRate(df3b, 'temp_Bin', target)
                #chisq2 = Chi2(df2b, 'total', 'bad', overallRate)
                chisq2 = Chi2(df2b, 'total', 'bad')
                if chisq1 < chisq2:
                    cutOffPoints.remove(cutOffPoints[currentIndex - 1])
                else:
                    cutOffPoints.remove(cutOffPoints[currentIndex])
            # 完成合并之后，需要再次计算新的分箱准则下，每箱是否同时包含好坏样本
            groupedvalues = df2['temp'].apply(lambda x: AssignBin(x, cutOffPoints))
            df2['temp_Bin'] = groupedvalues
            (binBadRate, regroup) = BinBadRate(df2, 'temp_Bin', target)
            [minBadRate, maxBadRate] = [min(binBadRate.values()), max(binBadRate.values())]
        # 需要检查分箱后的最小占比
        if minBinPcnt > 0:
            groupedvalues = df2['temp'].apply(lambda x: AssignBin(x, cutOffPoints))
            df2['temp_Bin'] = groupedvalues
            valueCounts = groupedvalues.value_counts().to_frame()
            valueCounts['pcnt'] = valueCounts['temp'].apply(lambda x: x * 1.0 / N)
            valueCounts = valueCounts.sort_index()
            minPcnt = min(valueCounts['pcnt'])
            while minPcnt < minBinPcnt and len(cutOffPoints) > 2:
                # 找出占比最小的箱
                indexForMinPcnt = valueCounts[valueCounts['pcnt'] == minPcnt].index.tolist()[0]
                # 如果占比最小的箱是最后一箱，则需要和上一个箱进行合并，也就意味着分裂点cutOffPoints中的最后一个需要移除
                if indexForMinPcnt == max(valueCounts.index):
                    cutOffPoints = cutOffPoints[:-1]
                # 如果占比最小的箱是第一箱，则需要和下一个箱进行合并，也就意味着分裂点cutOffPoints中的第一个需要移除
                elif indexForMinPcnt == min(valueCounts.index):
                    cutOffPoints = cutOffPoints[1:]
                # 如果占比最小的箱是中间的某一箱，则需要和前后中的一个箱进行合并，依据是较小的卡方值
                else:
                    # 和前一箱进行合并，并且计算卡方值
                    currentIndex = list(valueCounts.index).index(indexForMinPcnt)
                    prevIndex = list(valueCounts.index)[currentIndex - 1]
                    df3 = df2.loc[df2['temp_Bin'].isin([prevIndex, indexForMinPcnt])]
                    (binBadRate, df2b) = BinBadRate(df3, 'temp_Bin', target)
                    #chisq1 = Chi2(df2b, 'total', 'bad', overallRate)
                    chisq1 = Chi2(df2b, 'total', 'bad')
                    # 和后一箱进行合并，并且计算卡方值
                    laterIndex = list(valueCounts.index)[currentIndex + 1]
                    df3b = df2.loc[df2['temp_Bin'].isin([laterIndex, indexForMinPcnt])]
                    (binBadRate, df2b) = BinBadRate(df3b, 'temp_Bin', target)
                    #chisq2 = Chi2(df2b, 'total', 'bad', overallRate)
                    chisq2 = Chi2(df2b, 'total', 'bad')
                    if chisq1 < chisq2:
                        cutOffPoints.remove(cutOffPoints[currentIndex - 1])
                    else:
                        cutOffPoints.remove(cutOffPoints[currentIndex])
        cutOffPoints = special_attribute + cutOffPoints
        return cutOffPoints
    
def AssignBin(x, cutOffPoints,special_attribute=[]):
    '''
    :param x: 某个变量的某个取值
    :param cutOffPoints: 上述变量的分箱结果，用切分点表示
    :param special_attribute:  不参与分箱的特殊取值
    :return: 分箱后的对应的第几个箱，从0开始
    for example, if cutOffPoints = [10,20,30], if x = 7, return Bin 0. If x = 35, return Bin 3
    '''
    numBin = len(cutOffPoints) + 1 + len(special_attribute)
    if x in special_attribute:
        i = special_attribute.index(x)+1
        return 'Bin {}'.format(0-i)
    if x<=cutOffPoints[0]:
        return 'Bin 0'
    elif x > cutOffPoints[-1]:
        return 'Bin {}'.format(numBin-1)
    else:
        for i in range(0,numBin-1):
            if cutOffPoints[i] < x <=  cutOffPoints[i+1]:
                return 'Bin {}'.format(i+1)
    
def BadRateMonotone(df, sortByVar, target,special_attribute = []):
    '''
    :param df: 包含检验坏样本率的变量，和目标变量
    :param sortByVar: 需要检验坏样本率的变量
    :param target: 目标变量，0、1表示好、坏
    :param special_attribute: 不参与检验的特殊值
    :return: 坏样本率单调与否
    '''
    df2 = df.loc[~df[sortByVar].isin(special_attribute)]
    if len(set(df2[sortByVar])) <= 2:
        return True
    regroup = BinBadRate(df2, sortByVar, target)[1]
    combined = zip(regroup['total'],regroup['bad'])
    badRate = [x[1]*1.0/x[0] for x in combined]
    badRateNotMonotone = [badRate[i]<badRate[i+1] and badRate[i] < badRate[i-1] or badRate[i]>badRate[i+1] and badRate[i] > badRate[i-1]
                       for i in range(1,len(badRate)-1)]
    if True in badRateNotMonotone:
        return False
    else:
        return True
def SplitData(df, col, numOfSplit, special_attribute=[]):
    '''
    :param df: 按照col排序后的数据集
    :param col: 待分箱的变量
    :param numOfSplit: 切分的组别数
    :param special_attribute: 在切分数据集的时候，某些特殊值需要排除在外
    :return: 在原数据集上增加一列，把原始细粒度的col重新划分成粗粒度的值，便于分箱中的合并处理
    '''
    df2 = df.copy()
    if special_attribute != []:
        df2 = df.loc[~df[col].isin(special_attribute)]
    N = df2.shape[0]
    n = N%numOfSplit
    splitPointIndex = [i*n for i in range(1,numOfSplit)]
    rawValues = sorted(list(df2[col]))
    splitPoint = [rawValues[i] for i in splitPointIndex]
    splitPoint = sorted(list(set(splitPoint)))
    return splitPoint

def AssignGroup(x, bin):
    '''
    :param x: 某个变量的某个取值
    :param bin: 上述变量的分箱结果
    :return: x在分箱结果下的映射
    '''
    N = len(bin)
    if x<=min(bin):
        return min(bin)
    elif x>max(bin):
        return 10e10
    else:
        for i in range(N-1):
            if bin[i] < x <= bin[i+1]:
                return bin[i+1]



import os 
from .ChiMergeGroup import *
import numpy as np

def chimerge_clus(tmp1, col, y, max_interval,dropvalue_list=[]):
    tmp1=tmp1.drop(dropvalue_list,axis=1)
    tmp1['status']= y
    print("{} is in processing".format(col))
    cutOff = ChiMerge(tmp1, col, 'status', max_interval=max_interval,special_attribute=[],minBinPcnt=0)
    tmp1[col+'_Bin'] = tmp1[col].map(lambda x: AssignBin(x, cutOff,special_attribute=[]))
    return tmp1

def caliv(df,string,status):
    eps=0.00001
    crosstab=pd.crosstab(df[string],df[status]) + eps
    sample = df[status].value_counts() + eps
    frequence = crosstab/sample
    frequence['woe'] = np.log(frequence[1] / frequence[0])
    frequence['iv'] = (frequence[1] - frequence[0])*frequence['woe']
    frequence=frequence.reset_index(drop=False)
    var_dict=frequence.set_index(string)["woe"].to_dict()
    return {"iv":frequence['iv'].sum(),"woe":var_dict}

def char_group_chi(data_total,y,group_num,value_list=[]):
    '''
    Parameters
    ----------
    data_total :数据集
    value_list : 需要分组的字符变量名列表
    y : y变量名
    group_num : 分组数
    Returns 加了新的分组变量的数据集，分组映射字典,分组之后的新变量的变量名列表
    -------
    None.

    '''
    var_bin_list=[]
    var_dict_total={}
    iv_dict={}
    for col in value_list:
        data_total[col]=data_total[col].fillna("null")
        data=pd.merge(data_total.groupby([col])[y].sum().reset_index(drop=False),data_total.groupby([col])[y].count().reset_index(drop=False),how='inner',left_on=col,right_on=col)
        data["rate"]=data[y+"_x"]/data[y+"_y"]
        data=data.sort_values(by="rate")
        i=1
        dict_total={}
        for k in data[col]:
            dict1={k:i}
            dict_total=dict(dict_total,**dict1)
            i=i+1   
        data_total[col+"_group"]=data_total[col].map(lambda x :dict_total.get(x,0))

        # iv_dict={}
        # for m in [3,4,5,6,7]:
        #     data_total=chimerge_clus(data_total, col+"_group", train_data[y], max_interval=m,dropvalue_list=[])
        #     print(caliv(data_total,col+"_group_Bin",y))
        #     idict={str(m):caliv(data_total,col+"_group_Bin",y)} 
        #     iv_dict=dict(iv_dict,**idict)
        # print(iv_dict)
        # groupnum = max(iv_dict, key=iv_dict.get)   
        # print(groupnum)  
        data_total=chimerge_clus(data_total, col+"_group", data_total[y], max_interval=int(5),dropvalue_list=[])
        bin_data=data_total.drop_duplicates([col,col+"_group_Bin"],keep="last")[[col,col+"_group_Bin"]]
        var_dict=bin_data.set_index(col)[col+"_group_Bin"].to_dict()
        var_bin_list.append(col+"_group_Bin")
        vdict={col:{"bin":var_dict,"var_name":col+"_group_Bin"}}
        var_dict_total=dict(var_dict_total,**vdict)
        
        idict={col+"_group_Bin":caliv(data_total,col+"_group_Bin",y)} 
        iv_dict=dict(iv_dict,**idict)
    return {"data":data_total,"var_dict_total":var_dict_total,"var_bin_list":var_bin_list,"iv_dict":iv_dict}

def char_group_chi_main(data_total,id_name,y,group_num,drop_value=[]): 
    columns_data=data_total.dtypes.reset_index(drop=False)
    columns_list=list(columns_data[columns_data[0]=="object"]["index"])
    if id_name in columns_list or y in columns_list:
        columns_list.remove(id_name)
        columns_list.remove(y)
    else:
        pass
    for m in drop_value:
        columns_list.remove(m)
    char_value_bin=[]
    char_value=[]
    for k in columns_list:
        if len(data_total[k].unique())>5:
            char_value_bin.append(k)
        else:
            char_value.append(k)
    # print(char_value_bin)
    # print(char_value)
    gg=char_group_chi(data_total,y,5,value_list=char_value_bin)
    iv_dict={}
    for col in char_value:
        data=data_total[col].drop_duplicates([col])
        for k in data[col]:
            dict1={k:i}
            dict_total=dict(dict_total,**dict1)
            i=i+1   
        data_total[col+"_group"]=data_total[col].map(lambda x :dict_total.get(x,0))
        idict={col:caliv(data_total,col,y)} 
        iv_dict=dict(iv_dict,**idict)
    var_bin_list=gg.get("var_bin_list")
    if len(char_value)>0:
        var_bin_list.append(char_value)
    elif id_name in var_bin_list:
        var_bin_list.remove(id_name)
    else:
        pass
    # print(var_bin_list)
    iv_dict=dict(iv_dict,**gg.get("iv_dict"))
    return {"data":gg.get("data"),"var_dict_total":gg.get("var_dict_total"),"var_bin_list":var_bin_list,"iv_dict":iv_dict}

if __name__ == '__main__':
    train_data = pd.read_csv(r"F:\论文-风控\train.csv")  #读取训练集
    gg=char_group_chi_main(train_data,"id","isDefault",5,drop_value=["issueDate","earliesCreditLine"])
    

    



import pandas as pd
import os 
from .ChiMergeGroup import *
import numpy as np


def caliv(df,string,status):
    eps=0.00001
    crosstab=pd.crosstab(df[string],df[status]) + eps
    sample = df[status].value_counts() + eps
    frequence = crosstab/sample
    frequence['woe'] = np.log(frequence[1] / frequence[0])
    frequence['iv'] = (frequence[1] - frequence[0])*frequence['woe']
    frequence=frequence.reset_index(drop=False)
    var_dict=frequence.set_index(string)["woe"].to_dict()
    return {"iv":frequence['iv'].sum(),"woe":var_dict}
def chimerge_clus(tmp1, col, y, max_interval,dropvalue_list=[]):
    tmp1=tmp1.drop(dropvalue_list,axis=1)
    tmp1['status']= y
    print("{} is in processing".format(col))
    if -1 not in set(tmp1[col]):   
        #－1会当成特殊值处理。如果没有－1，则所有取值都参与分箱
        cutOff = ChiMerge(tmp1, col, 'status', max_interval=max_interval,special_attribute=[],minBinPcnt=0)
        tmp1[col+'_Bin'] = tmp1[col].map(lambda x: AssignBin(x, cutOff,special_attribute=[]))
        monotone = BadRateMonotone(tmp1, col+'_Bin', 'status')   # 检验分箱后的单调性是否满足
        while(not monotone):
            # 检验分箱后的单调性是否满足。如果不满足，则缩减分箱的个数。
            max_interval -= 1
            cutOff = ChiMerge(tmp1, col, 'status', max_interval=max_interval, special_attribute=[],
                                            minBinPcnt=0)
            tmp1[col + '_Bin'] = tmp1[col].map(lambda x: AssignBin(x, cutOff, special_attribute=[]))
            if max_interval == 2:
                # 当分箱数为2时，必然单调
                break
            monotone = BadRateMonotone(tmp1, col + '_Bin', 'status')
        newVar = col + '_Bin'
        tmp1[newVar] = tmp1[col].map(lambda x: AssignBin(x, cutOff, special_attribute=[]))
    else:
        max_interval = 5
        # 如果有－1，则除去－1后，其他取值参与分箱
        cutOff = ChiMerge(tmp1, col, 'status', max_interval=max_interval, special_attribute=[-1],
                                        minBinPcnt=0)
        tmp1[col + '_Bin'] = tmp1[col].map(lambda x: AssignBin(x, cutOff, special_attribute=[-1]))
        monotone = BadRateMonotone(tmp1, col + '_Bin', 'status',['Bin -1'])
        while (not monotone):
            max_interval -= 1
            # 如果有－1，－1的bad rate不参与单调性检验
            cutOff = sf.ChiMerge(tmp1, col, 'status', max_interval=max_interval, special_attribute=[-1],minBinPcnt=0)
            tmp1[col + '_Bin'] = tmp1[col].map(lambda x: AssignBin(x, cutOff, special_attribute=[-1]))
            if max_interval == 3:
                # 当分箱数为3-1=2时，必然单调
                break
            monotone = BadRateMonotone(tmp1, col + '_Bin', 'status',['Bin -1'])
            newVar = col + '_Bin'
            tmp1[newVar] = tmp1[col].map(lambda x: AssignBin(x, cutOff, special_attribute=[-1]))
    return tmp1

def group_class(initial_group_num,num_bins):
    
    for i in range(initial_group_num):
        #如果第一个组没有包含正样本或负样本，向后合并
        if 0 in num_bins[0][2:]:
            num_bins[0:2] = [(
                num_bins[0][0],
                num_bins[1][1],
                num_bins[0][2]+num_bins[1][2],
                num_bins[0][3]+num_bins[1][3])]
            continue
    
        """
        如果原本的第一组和第二组都没有包含正样本，或者都没有包含负样本，那即便合并之后，第一行的组也还是没有
        包含两种样本
        所以我们在每次合并完毕之后，还需要再检查，第一组是否已经包含了两种样本
        这里使用continue跳出了本次循环，开始下一次循环，所以回到了最开始的for i in range(20), 让i+1
        这就跳过了下面的代码，又从头开始检查，第一组是否包含了两种样本
        如果第一组中依然没有包含两种样本，则if通过，继续合并，每合并一次就会循环检查一次，最多合并20次
        如果第一组中已经包含两种样本，则if不通过，就开始执行下面的代码
        """
        #已经确认第一组中肯定包含两种样本了，如果其他组没有包含两种样本，就向前合并
        #此时的num_bins已经被上面的代码处理过，可能被合并过，也可能没有被合并
        #但无论如何，我们要在num_bins中遍历，所以写成in range(len(num_bins))
        for i in range(len(num_bins)):
            if 0 in num_bins[i][2:]:
                num_bins[i-1:i+1] = [(
                    num_bins[i-1][0],
                    num_bins[i][1],
                    num_bins[i-1][2]+num_bins[i][2],
                    num_bins[i-1][3]+num_bins[i][3])]
            break
            #如果对第一组和对后面所有组的判断中，都没有进入if去合并，则提前结束所有的循环
        else:
            break
        return num_bins
def num_group_chi(data_total,y,initial_group_num,group_num,value_list=[]):
    var_bin_list=[]
    var_dict_total={}
    iv_dict={}
    for col in value_list:
        data_total[col]=data_total[col].fillna(-998)
        data_total["qcut"], updown = pd.qcut(data_total[col], retbins=True, q=int(initial_group_num),duplicates="drop")#等频分箱
        coount_y0 = data_total[data_total[y] == 0].groupby(by="qcut").count()[y]
        coount_y1 = data_total[data_total[y] == 1].groupby(by="qcut").count()[y]
        num_bins = [*zip(updown,updown[1:],coount_y0,coount_y1)]
        num_bins=group_class(initial_group_num,num_bins)
        gg=list([k[1] for k in num_bins])
        gg.append(-9999.0)
        gg.sort()
        data_total[col+"_group"]=pd.cut(data_total[col], gg)
        data_total=data_total.sort_values(by=col)
        data=data_total.drop_duplicates([col+"_group"])
        i=1
        dict_total={}
        for k in data[col+"_group"]:
            dict1={str(k):i}
            dict_total=dict(dict_total,**dict1)
            i=i+1  
        data_total[col+"_group"]=data_total[col+"_group"].map(lambda x :dict_total.get(str(x),0))
        data_total=chimerge_clus(data_total, col+"_group", data_total[y], max_interval=group_num,dropvalue_list=[])
        total_dict={}
        list_group=list(set(data_total[col+"_group_Bin"]))
        list_group.sort()
        for m in range(0,len(list_group)):
            h="Bin "+str(m)
            if h=="Bin 0":
                d_tatal=data_total[data_total[col+"_group_Bin"]==h]
                v_dict={h:[-9999.0,max(d_tatal[col])]}
            elif h=="Bin "+str(len(list_group)-1):
                d_tatal=data_total[data_total[col+"_group_Bin"]==h]
                v_dict={h:[min_1,10000000000]}                
            else:
                d_tatal=data_total[data_total[col+"_group_Bin"]==h]
                v_dict={h:[min_1,max(d_tatal[col])]}
            min_1=max(d_tatal[col])
            total_dict=dict(total_dict,**v_dict)
        # print(total_dict)
        v_total_dict={col:{"bin":total_dict,"value_name":col+"_group_Bin"}}
        var_dict_total=dict(var_dict_total,**v_total_dict)
        idict={col+"_group_Bin":caliv(data_total,col+"_group_Bin",y)} 
        iv_dict=dict(iv_dict,**idict)
        var_bin_list.append(col+"_group_Bin")
    return {"data":data_total,"var_dict_total":var_dict_total,"var_bin_list":var_bin_list,"iv_dict":iv_dict}

def num_group_chi_main(train_data,id_name,y,initial_group_num,group_num): 
    columns_data=train_data.dtypes.reset_index(drop=False)
    columns_list=list(columns_data[(columns_data[0]=="int64")|(columns_data[0]=="float64")|(columns_data[0]=="int32")|(columns_data[0]=="float32")]["index"])
    columns_list.remove(id_name)
    columns_list.remove(y)
    num_value_bin=[]
    num_value=[]
    for k in columns_list:
        if len(train_data[k].unique())>5:
            num_value_bin.append(k)
        else:
            num_value.append(k)
    gg=num_group_chi(train_data,y,20,5,value_list=num_value_bin)
    iv_dict={}
    for col in num_value:
        idict={col:caliv(train_data,col,y)} 
        iv_dict=dict(iv_dict,**idict)
    var_bin_list=gg.get("var_bin_list")
    for m in num_value:
        # print(m)
        var_bin_list.append(m)
    iv_dict=dict(iv_dict,**gg.get("iv_dict"))
    if id_name in var_bin_list:
        var_bin_list.remove(id_name)
    return {"data":gg.get("data"),"var_dict_total":gg.get("var_dict_total"),"var_bin_list":var_bin_list,"iv_dict":iv_dict}
if __name__ == '__main__':
    train_data = pd.read_csv(r"F:\论文-风控\train.csv")  #读取训练集
    # gg=num_group_chi_main(train_data,"id","isDefault",20,5)










from scipy.stats import scoreatpercentile
import seaborn as sns
import matplotlib.pyplot as plt
import pandas as pd 
from sklearn.metrics import roc_curve, auc
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split
import pandas as pd
from .NumBestGroup import *
from .CharBestGroup import *
import statsmodels.api as sm
class MakeModel(object):
    def __init__(self,data,id_name,y_name,iv_num,group_num,drop_value=[]):
        self.train_data,self.val_data=train_test_split(data, test_size=0.3, random_state=545)
        self.id_name=id_name
        self.y_name=y_name
        self.drop_value=drop_value
        self.iv_num=iv_num
        self.group_num=group_num
    
    
    
    def select_value_iv(self):
        '''
        返回数值变量 字符变量中大于设定的iv_num的变量列表  
        '''
        feature_list = [k for k,v in self.num_object.get("iv_dict").items() if v.get("iv")>self.iv_num ]
        feature_list1 = [k for k,v in self.char_object.get("iv_dict").items() if v.get("iv")>self.iv_num ]
        for j in feature_list1:
            feature_list.append(j)
        self.feature_list=feature_list
    
    def corr_x(self):
        return X.corr()
    
    def ks_calc_auc(self,data,pred,y_label):
      '''
      功能: 计算KS值，输出对应分割点和累计分布函数曲线图
      输入值:
      data: 二维数组或dataframe，包括模型得分和真实的标签
      pred: 一维数组或series，代表模型得分（一般为预测正类的概率）
      y_label: 一维数组或series，代表真实的标签（{0,1}或{-1,1}）
      输出值:
      'ks': KS值
      '''
      fpr,tpr,thresholds= roc_curve(data[y_label],data[pred])
      ks = max(tpr-fpr)
      return ks
    

    
    def bin_woe(self,n,x):   
        var=n.replace("_group_Bin","")
        for b in self.num_object.get("var_dict_total").get(var).get("bin"):
            # print(b)
            if x>self.num_object.get("var_dict_total").get(var).get("bin").get(b)[0]  and x<=self.num_object.get("var_dict_total").get(var).get("bin").get(b)[1]:
                return self.num_object.get("iv_dict").get(n).get("woe").get(b)
            else:
                continue  
            
    def deal_raw_data(self,data):
        for n in self.num_object.get("var_bin_list"):
            print(n)
            if "_group_Bin" in n:
                var=n.replace("_group_Bin","")
                data[var]=data[var].fillna(-998)
                data[self.num_object.get("var_dict_total").get(var).get("value_name")]= data[var].map(lambda x :self.bin_woe(n,x))
            else:
                data[n]= data[n].map(lambda x :self.num_object.get("iv_dict").get(n).get("woe").get(x))
        for n in self.char_object.get("var_bin_list"):
            if "_group_Bin" in n:
                n=n.replace("_group_Bin","")
                data[n]=data[n].fillna("null")
                b=self.char_object.get("var_dict_total").get(n).get("var_name")
                data[self.char_object.get("var_dict_total").get(n).get("var_name")]= data[n].map(lambda x :self.char_object.get("iv_dict").get(b).get("woe").get(self.char_object.get("var_dict_total").get(n).get("bin").get(x)))
            else:
                data[n]= data[n].map(lambda x :self.char_object.get("iv_dict").get(n).get("woe").get(x))
        return data

    def model_main(self,feature_list=[]):     
        self.num_object=num_group_chi_main(self.train_data,self.id_name,self.y_name,20,self.group_num)
        self.char_object=char_group_chi_main(self.train_data,self.id_name,self.y_name,5,self.drop_value)
        # num_object_data=self.woe_change(self.num_object)
        # char_object_data=self.woe_change(self.char_object)   
        self.select_value_iv()
        if len(feature_list)==0:
            feature_list=self.feature_list
        else:
            feature_list=feature_list
        # 处理原始训练集
        self.train_data=self.deal_raw_data(self.train_data)
        y = self.train_data[self.y_name]
        X = self.train_data[feature_list]
        X['intercept'] = [1]*X.shape[0]
        LR = sm.Logit(y, X).fit()
        summary = LR.summary()
        pvals = LR.pvalues
        pvals = pvals.to_dict()
        self.train_data['predicted'] = LR.predict(X.astype(float))
        # 处理原始测试集
        self.val_data=self.deal_raw_data(self.val_data)
        X_val = self.val_data[self.feature_list]
        X_val['intercept'] = [1]*X_val.shape[0]
        self.val_data['predicted']= LR.predict(X_val.astype(float))
        ks=self.ks_calc_auc(self.train_data,'predicted', self.y_name)
        auc=roc_auc_score(list(self.train_data[self.y_name]),list(self.train_data['predicted']))
        val_ks=self.ks_calc_auc(self.val_data,'predicted', self.y_name)
        val_auc=roc_auc_score(list(self.val_data[self.y_name]),list(self.val_data['predicted']))        
        self.result={"model_summary":summary,"ks":ks,"auc":auc,"val_ks":val_ks,"val_auc":val_auc}        
if __name__ == '__main__':
    data = pd.read_csv(r"F:\论文-风控\train.csv")  #读取训练集
    m=MakeModel(data,"id","isDefault",0.02,5,drop_value=["issueDate","earliesCreditLine"])
    m.model_main()
    print(m)

    
    
    
    

